#
# Makefile for the kernel SRAM block device drivers.
#
# Advantech eAutomation Division
#


UNAME := $(shell uname -r)
PWD := $(shell pwd)
ADVMOD := advsram

# FSTYPE 
# 1. Leave it empty
#    It can be failed if SRAM does not contain a file system.
# 2. File system types, such as ext2
#    It will force to make a specified file system.
#    Everything in SRAM will be destroied.
# FSTYPE :=
# FSTYPE := ext2
# FSTYPE := vfat
FSTYPE := ext2
ifneq ($(FSTYPE),) 
MKFS := @mkfs -t $(FSTYPE) /dev/$(ADVMOD) > /dev/null
endif
ifeq "$(UNAME)" "2.4.27-2-386"
LINUX_SRC=/usr/src/kernel-headers-2.4.27-2-386
else 
LINUX_SRC=	/usr/src/linux-2.4
endif
ARCH_FLAGS=    $(I386_FLAGS)

# Define the advsram driver options as appropriate .
INCL_FLAGS=  -I. -I$(LINUX_SRC)/include
DEF_FLAGS=   -D__KERNEL__ -DLINUX -DLOCAL_HEADERS -DEXPORT_SYMTAB 
I386_FLAGS=  -march=i686 -O2 -Wall 
CFLAGS=	     $(ARCH_FLAGS) $(INCL_FLAGS) $(DEF_FLAGS) -DMODULE
CC=          gcc

default: advsram.o
	ld -r  sram.o advsram_pci.o -o advsram.o

	
advsram.o:sram.o advsram_pci.o
	

install:
	$(shell if grep $(ADVMOD) /proc/mounts > /dev/null ; then \
	umount /mnt/$(ADVMOD) ; fi)
	$(shell if grep $(ADVMOD) /proc/modules > /dev/null ; then \
	rmmod $(ADVMOD) ; fi)
	$(shell [ -b /dev/$(ADVMOD) ] && rm -f /dev/$(ADVMOD))
	@insmod $(ADVMOD).o
	@mknod /dev/$(ADVMOD) b $$(awk '$$2=="advsram" {print $$1}' /proc/devices ) 0
	@chmod a+wx /dev/$(ADVMOD)	


testraw: testraw.c
	@$(CC) -o testraw testraw.c

testdirect:testdirect.c
	@$(CC) -o testdirect testdirect.c

fs: install 
	$(shell [ -d /mnt/$(ADVMOD) ] && \
		rmdir --ignore-fail-on-non-empty /mnt/$(ADVMOD))	
	$(MKFS)
	@mkdir /mnt/$(ADVMOD)
	@mount /dev/$(ADVMOD) /mnt/$(ADVMOD)
	
raw: install testraw       
	@raw /dev/raw/raw1 /dev/$(ADVMOD)
	@./testraw /dev/raw/raw1 /dev/advsram

direct: install testdirect
	@chmod a+wx /dev/$(ADVMOD)
	@./testdirect /dev/advsram
	
uninstall:
	$(shell if grep $(ADVMOD) /proc/mounts > /dev/null ; then \
	 umount /mnt/$(ADVMOD) ; fi)
	$(shell if grep $(ADVMOD) /proc/modules > /dev/null ; then \
	 rmmod $(ADVMOD) ; fi)
	$(shell [ -b /dev/$(ADVMOD) ] && rm -f /dev/$(ADVMOD))
	$(shell [ -d /mnt/$(ADVMOD) ] && \
		rmdir --ignore-fail-on-non-empty /mnt/$(ADVMOD))	

clean:
	@rm -f *.o
	@rm -f testraw
	@rm -f testdirect
