#
# Makefile for the kernel SRAM block device drivers.
#
# Advantech eAutomation Division
#
MAJOR :=
UNAME := $(shell uname -r)
PWD := $(shell pwd)
ADVMOD := advsram
CC := gcc 
# FSTYPE 
# 1. Leave it empty
#    It can be failed if SRAM does not contain a file system.
# 2. File system types, such as ext2
#    It will force to make a specified file system.
#    Everything in SRAM will be destroied.
# FSTYPE :=
# FSTYPE := ext2
# FSTYPE := vfat
FSTYPE := ext2
ifneq ($(FSTYPE),) 
MKFS := @mkfs -t $(FSTYPE) /dev/$(ADVMOD) > /dev/null
endif

ifneq ($(KERNELRELEASE),)
obj-m := advsram.o
advsram-objs	:= sram.o	\
	    advsram_pci.o		 
else
	KDIR := /lib/modules/$(shell uname -r)/build/
	PWD  := $(shell pwd)
all:
	@echo "This driver support advantech sram device"
	$(MAKE) -w -C $(KDIR) SUBDIRS=$(PWD) modules
endif


install:
	$(shell if grep $(ADVMOD) /proc/mounts > /dev/null ; then \
	umount /mnt/$(ADVMOD) ; fi)
	$(shell if grep $(ADVMOD) /proc/modules > /dev/null ; then \
	rmmod $(ADVMOD) ; fi)
	$(shell [ -b /dev/$(ADVMOD) ] && rm -f /dev/$(ADVMOD))	
	@insmod $(ADVMOD).ko
	@sleep 2
	-@rm -f /dev/$(ADVMOD)	
	@mknod /dev/$(ADVMOD) b $$(awk '$$2=="advsram" {print $$1}' /proc/devices ) 0
	@sleep 1
	@chmod a+wx /dev/$(ADVMOD)	

testraw: testraw.c
	@$(CC) -o testraw testraw.c

testdirect:testdirect.c
	@$(CC) -o testdirect testdirect.c

	
fs: install
	$(shell [ -d /mnt/$(ADVMOD) ] && \
		rmdir --ignore-fail-on-non-empty /mnt/$(ADVMOD))	
	$(MKFS)
	@mkdir /mnt/$(ADVMOD)
	@mount /dev/$(ADVMOD) /mnt/$(ADVMOD)
	
raw: install testraw        
	@raw /dev/raw/raw1 /dev/$(ADVMOD)
	@./testraw /dev/raw/raw1 /dev/advsram
	
direct: install testdirect       
	@./testdirect /dev/advsram
		
uninstall:
	$(shell if grep $(ADVMOD) /proc/mounts > /dev/null ; then \
	 umount /mnt/$(ADVMOD) ; fi)
	$(shell if grep $(ADVMOD) /proc/modules > /dev/null ; then \
	 rmmod $(ADVMOD) ; fi)
	$(shell [ -b /dev/$(ADVMOD) ] && rm -f /dev/$(ADVMOD))
	$(shell [ -d /mnt/$(ADVMOD) ] && \
		rmdir --ignore-fail-on-non-empty /mnt/$(ADVMOD))	
	@echo ""

clean:
	-@rm -f tags
	-@rm -f *.o *.ko
	-@rm -f .*.cmd *.mod.c
	-@rm -rf .tmp_versions
	-@rm -rf *~
	-@rm -rf *odule*
	-@rm -rf testdirect

