// #############################################################################
// *****************************************************************************
//                  Copyright ( c ) 2009, Advantech Automation Corp.
//      THIS IS AN UNPUBLISHED WORK CONTAINING CONFIDENTIAL AND PROPRIETARY
//               INFORMATION WHICH IS THE PROPERTY OF ADVANTECH AUTOMATION CORP.
//
//    ANY DISCLOSURE, USE, OR REPRODUCTION, WITHOUT WRITTEN AUTHORIZATION FROM
//               ADVANTECH AUTOMATION CORP., IS STRICTLY PROHIBITED.
// *****************************************************************************
// #############################################################################
/***
 * File:              advsram_pci.c
 * Author:            James Dai
 * Created:           2009-8-20
 * Revision:          1.00 
 *
 * Description:
 *
 * Driver init section for advantech PCI SRAM device
 *
 */


#include <linux/pci.h>
#include "advsram.h"

#define ADVANTECH_VANDORID 0x13fe
extern int    i_iobase;
extern int    i_iorange;
static struct pci_device_id advsram_board_table[] = {
	{ADVANTECH_VANDORID, 0x1011, PCI_ANY_ID, PCI_ANY_ID, 0, 0, PCI_ANY_ID},
	{0,}			
};

static int __init advsram_init_one(struct pci_dev *pdev, const struct pci_device_id *id);
static void __exit advsram_remove_one(struct pci_dev *dev);

static struct pci_driver advsram_driver = {
	name:		"advsram",
	id_table:	advsram_board_table,
	probe:		advsram_init_one,
	remove:		advsram_remove_one,
};



static int __init advsram_init_one(struct pci_dev *pdev, const struct pci_device_id *id ) 
{
   unsigned int address,address_len;
   if (pci_enable_device (pdev))
   {
      return -EIO;
   }

#if 0
   printk("Get a pci can board at bus: %d, slot: %d, vendor: %x, device: %x\n", 
      pdev->bus->number,
      pdev->devfn, 
      pdev->vendor,
      pdev->device);
#endif
   address = pci_resource_start(pdev, 0);
   address_len = (pci_resource_len(pdev, 0));
   if ( request_mem_region(address, address_len , "advsram") == NULL ) 
   {
      printk ("Device I/O 0x%X is not free.\n",address);
      goto out_iobase0;
   }
   
   advdev = pdev;
   AdvSRAMInfo.ul_BankEnableAddress    = DEFAULT_BANK_ENABLE_ADDRESS;
   AdvSRAMInfo.ul_BankLength           = address_len;
   AdvSRAMInfo.ul_BankMemoryAddress    = address;
   AdvSRAMInfo.ul_BankNumber           = 1;
   AdvSRAMInfo.ul_BankNumberAddress    = DEFAULT_BANK_NUMBER_ADDRESS;
   AdvSRAMInfo.ul_BankSwitchAddress    = DEFAULT_BANK_SWITCH_ADDRESS;
   AdvSRAMInfo.ul_ManufactureIDAddress = DEFAULT_MANUFACTUREID_ADDRESS;
   
   i_iobase = pci_resource_start(pdev, 1);
   i_iorange = pci_resource_len(pdev, 1);
   request_region(i_iobase, i_iorange, "advsram");
   printk("FPGA Version:%04x\n",(inb(i_iobase+1) << 8) + inb(i_iobase) );
   return 0;
out_iobase0:
	release_mem_region(address, address_len);
	return -EIO;
}

static void __exit advsram_remove_one( struct pci_dev *dev) //zdd modified
{
   release_mem_region(AdvSRAMInfo.ul_BankMemoryAddress, AdvSRAMInfo.ul_BankLength);
}

int init_advsram_pci(void)
{
   int ret;
   ret = 0;
   ret = pci_register_driver(&advsram_driver);
   if(ret < 0) 
   {
      return -ENODEV;
   }
   return 0;
}

void exit_advsram_pci(void)
{
   pci_unregister_driver(&advsram_driver);
}
