//*****************************************************************************
//
//                 Copyright (c) 2011 Advantech Co., Ltd.
//
//                         Linux SRAMDISK device driver
//
//      This is An Unpublished Work Containing Confidential  And Proprietary
//         Information Which Is The Property Of Advantech  Automation Corp.
//
//    Any Disclosure, Use, Or Reproduction, Without Written Authorization From
//               Advantech Automation Corp., Is Strictly Prohibit.
//
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// File:        advsram.c
// Created:     2008-9-12
// Description: The driver is designed to access the SRAMDISK for UNO-2160/3062,
//              and has filesystems created on it and mounted just like a 
//              regular block device.
// 
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for any purpose and without fee is hereby granted, provided
// that the above copyright notice appear in all copies and that both that
// copyright notice and this permission notice appear in supporting
// documentation.  This software is provided "as is" without express or
// implied warranty.
//=============================================================================

#define ADVANTECH
#define ADVANTECH_INFO
//#define ADVANTECH_DEBUG

#define ADVANTECH_SRAM_VER	"2.16"
#define ADVANTECH_SRAM_DATE	"03/26/2011"

#include <linux/init.h>
#include <linux/module.h>
#include <linux/ioport.h>
#include <linux/hdreg.h>  /* HDIO_GETGEO */

#include <asm/uaccess.h>
#include <asm/io.h>

#include "advsram.h"

#define MAJOR_NR advsd_major        // force definitions on in blk.h
int     advsd_major;                // must be declared before including blk.h

#define TRUE                  (1)
#define FALSE                 (0)

#define DEVICE_NR(device) MINOR(device)   
#define DEVICE_NAME "advsram"       // name for messaging: Advantech SRAMDISK
#define DEVICE_INTR adsd_intrptr    // pointer to the bottom half
#define DEVICE_NO_RANDOM            // no entropy to contribute

#include <linux/version.h>
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
#include <linux/blkdev.h>
#else
#include <linux/blk.h>
#endif
#include <linux/blkpg.h>

#ifdef ADVANTECH_DEBUG
#define DEBUGPRINT	printk
#else
#define DEBUGPRINT(a, ...)
#endif

#ifdef ADVANTECH_INFO
#define INFOPRINT	printk
#else
#define INFOPRINT(a, ...)
#endif

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
static struct gendisk *rd_disk;
static struct request_queue *rd_queue;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,31)
static spinlock_t sram_lock = SPIN_LOCK_UNLOCKED;
#else
static DEFINE_SPINLOCK(sram_lock);
#endif
#else
static int     asd_hardsec[ NUM_SRAMDISKS ];      // Size of real blocks in byte
static int     asd_blocksizes[ NUM_SRAMDISKS ];   // Size of 1024 byte blocks
static int     asd_sizes[ NUM_SRAMDISKS ];        // Size of the device in bytes
#endif

static unsigned char * uc_ptaddr;		  // pointer to the remap space 
static int    current_device = -1;
int    i_iobase;
int    i_iorange;
static int  is_pci = 0;
struct pci_dev *advdev;
DISK_INFO     AdvSRAMInfo = {0};
extern int init_advsram_pci(void);
extern void exit_advsram_pci(void);
#if defined (MODULE)
MODULE_LICENSE("GPL");
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,23)
static void do_asd_request( struct request_queue * q )
#else
static void do_asd_request( request_queue_t * q )
#endif
{
	int i, i_StartBank, i_EndBank;
	u_long start, len;//, addr, size;
	u_long ul_StartBank, ul_EndBank, ul_DiskOffset, ul_BufOffset, ul_Size;
	unsigned char*  uc_ptDisk; 

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)	  
	struct request *req;
#endif
	
	DEBUGPRINT( "----->>>>> DO_ASD_REQUEST! \n");
	i = i_StartBank = i_EndBank = 0;   
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,31)
	req = blk_fetch_request(q);
	while ( req )
	{
		int rw = rq_data_dir(req);
		DEBUGPRINT( DEVICE_NAME": request %p: dir %d sector %ld (nr. %u) \n", 
			req, rw, (unsigned long) blk_rq_pos(req), req->current_nr_sectors );
		// per sector = 512 Bytes
		start = blk_rq_pos(req) << 9;    
		// the number of the sectors to transfer, per sector = 512 Bytes
		len  = blk_rq_cur_bytes(req); 
		int err = 0; 
#else
	while ( (req = elv_next_request(q)) != NULL )
	{
		int rw = rq_data_dir(req);
		DEBUGPRINT( DEVICE_NAME": request %p: dir %d sector %ld (nr. %u) \n", 
			req, rw, (unsigned long) req->sector, req->current_nr_sectors );
		// per sector = 512 Bytes
		start = req->sector << 9;    
		// the number of the sectors to transfer, per sector = 512 Bytes
		len  = req->current_nr_sectors << 9;  
#endif
		DEBUGPRINT( DEVICE_NAME": Start addr:%ld, Length:%ld\n", start, len );   
		
		ul_StartBank = start / AdvSRAMInfo.ul_BankLength;
		ul_EndBank   = ( start + len - 1) / AdvSRAMInfo.ul_BankLength;
		DEBUGPRINT( DEVICE_NAME": Start Bank: %ld, End Bank: %ld\n", ul_StartBank, ul_EndBank );
		
		if ( ( start + len ) > AdvSRAMInfo.ul_DiskSize )
		{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,31)
			INFOPRINT( DEVICE_NAME ": bad access: block=%ld, count=%u\n",
			(long int)blk_rq_pos(req),blk_rq_cur_sectors(req));
			err = -EIO;
			goto done;
#else
			INFOPRINT( DEVICE_NAME ": bad access: block=%ld, count=%u\n",
			(unsigned long) req->sector,
			req->current_nr_sectors);
			INFOPRINT( DEVICE_NAME ": request past end of device\n");	
			end_request(req, FALSE );
			continue;
#endif
			
		}	
		
		if ( ( rw != READ ) && ( rw != WRITE ) )
		{
			//INFOPRINT( DEVICE_NAME ": unknow command: %d\n", rq_data_dir(req) );
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,31)
			err = -EIO;
			goto done;
#else
			end_request(req, FALSE );
#endif
			continue;
		}
		
		ul_BufOffset  = 0;
		ul_DiskOffset = start % AdvSRAMInfo.ul_BankLength;
		ul_Size       = ( ul_DiskOffset + len <= AdvSRAMInfo.ul_BankLength )? 
				len : (AdvSRAMInfo.ul_BankLength - ul_DiskOffset );
		for ( i = ul_StartBank; i <= ul_EndBank; i++ )
		{
			uc_ptDisk = uc_ptaddr + ul_DiskOffset;
			DEBUGPRINT( DEVICE_NAME": R/W - bank = %d, ul_Size = %ld\n", i, ul_Size );
			
			// set bank
			if ( is_pci == 0 )
			{
				outb( i, AdvSRAMInfo.ul_BankSwitchAddress);
			}
			if ( rw == READ )
			{  
				DEBUGPRINT(DEVICE_NAME": READ-BANK = %d\n", i );
				memcpy( &req->buffer[ul_BufOffset], uc_ptDisk, ul_Size );            
			}
			else if ( rw == WRITE )
			{
				DEBUGPRINT(DEVICE_NAME": WRITE-BANK = %d\n", i );
				if ( is_pci == 0 )
				{
					outb( 1, AdvSRAMInfo.ul_BankEnableAddress);
				}
				memcpy( uc_ptDisk, &req->buffer[ul_BufOffset], ul_Size );
				if ( is_pci == 0 )
				{
					outb( 0, AdvSRAMInfo.ul_BankEnableAddress);
				}
			}
			
			if ( i < ul_EndBank )
			{
				ul_DiskOffset = 0;
				ul_BufOffset += ul_Size;
				ul_Size       = len - ul_BufOffset;
				if ( ul_Size > AdvSRAMInfo.ul_BankLength )
					ul_Size = AdvSRAMInfo.ul_BankLength;
			}         
		} 
		if ( is_pci == 0 )
		{
			outb( 0, AdvSRAMInfo.ul_BankSwitchAddress);
		}  
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,31)	
	done:
		if (!__blk_end_request_cur(req, err))
			req = blk_fetch_request(q);
#else   
		end_request(req, TRUE );
#endif
	}
#else
	while ( TRUE )
	{
		INIT_REQUEST;
		DEBUGPRINT( DEVICE_NAME": request %p: cmd %i sector %li (nr. %li) \n", 
		CURRENT, CURRENT->cmd, CURRENT->sector, CURRENT->current_nr_sectors );
		// check if the minor number is in range
		if ( DEVICE_NR( CURRENT->rq_dev ) > NUM_SRAMDISKS )
		{
			INFOPRINT( DEVICE_NAME": request for unknown device.\n");
			end_request( 0 );
			continue;
		}

		// per sector = 512 Bytes
		start = CURRENT->sector << 9;    
		// the number of the sectors to transfer, per sector = 512 Bytes
		len  = CURRENT->current_nr_sectors << 9;  
		DEBUGPRINT(DEVICE_NAME": CURRENT->sector:%ld, current_nr_sectors:%ld, nr_sectors:%ld\n", 
		CURRENT->sector, CURRENT->current_nr_sectors, CURRENT->nr_sectors );

		ul_StartBank = start / AdvSRAMInfo.ul_BankLength;
		ul_EndBank   = ( start + len ) / AdvSRAMInfo.ul_BankLength;
		DEBUGPRINT( DEVICE_NAME": Start Bank: %ld, End Bank: %ld\n", 
		ul_StartBank, ul_EndBank );
      
		if ( ( start + len ) > AdvSRAMInfo.ul_DiskSize )
		{
			INFOPRINT( DEVICE_NAME ": bad access: block=%ld, count=%ld\n",
			CURRENT->sector,
			CURRENT->current_nr_sectors);
			INFOPRINT( DEVICE_NAME " : request past end of device\n");
			end_request( FALSE );
			continue;
		}	
      
		if ( ( CURRENT->cmd != READ ) && ( CURRENT->cmd != WRITE ) )
		{
			INFOPRINT( DEVICE_NAME ": bad command: %d\n", CURRENT->cmd );

			end_request( FALSE );

			continue;
		}

		ul_BufOffset  = 0;
		ul_DiskOffset = start % AdvSRAMInfo.ul_BankLength;
		ul_Size       = ( ul_DiskOffset + len <= AdvSRAMInfo.ul_BankLength )? len : (AdvSRAMInfo.ul_BankLength - ul_DiskOffset );
      
		for ( i = ul_StartBank; i <= ul_EndBank; i++ )
		{
			uc_ptDisk = uc_ptaddr + ul_DiskOffset;
			// set bank
			if ( is_pci == 0 )
			{
				outb( i, AdvSRAMInfo.ul_BankSwitchAddress );
			}
			if ( CURRENT->cmd == READ )
			{  
				DEBUGPRINT(DEVICE_NAME": READ-BANK = %d\n", i );
				memcpy( CURRENT->buffer, uc_ptDisk, ul_Size );            
			}
			else if ( CURRENT->cmd == WRITE )
			{
				DEBUGPRINT( DEVICE_NAME": WRITE-BANK = %d\n", i );
				if ( is_pci == 0 )
				{
					outb( 1, AdvSRAMInfo.ul_BankEnableAddress);
				}
				memcpy( uc_ptDisk, CURRENT->buffer, ul_Size );
				if ( is_pci == 0 )
				{
					outb( 0, AdvSRAMInfo.ul_BankEnableAddress);
				}
			}
      
			if ( i < ul_EndBank )
			{
				ul_DiskOffset = 0;
				ul_BufOffset += ul_Size;
				ul_Size       = len - ul_BufOffset;
				if ( ul_Size > AdvSRAMInfo.ul_BankLength )
				{
					ul_Size    = AdvSRAMInfo.ul_BankLength;
				}         
			}         
		} 
		if ( is_pci == 0 )
		{
			outb( 0, AdvSRAMInfo.ul_BankSwitchAddress);    
		}  

		end_request( TRUE );
	}
#endif   	
}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,31)	
static int advsram_open ( struct inode *inode, struct file *filp )
#else
static int advsram_open(struct block_device *bdev, fmode_t mode)
#endif
{
	int device;    
	int rc = -ENODEV;    
	unsigned char uc_temp;
	unsigned long size, paddr;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)	
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,31)	
	device = iminor(inode); 
#else  
   	device = MINOR(bdev->bd_dev);
#endif
	if ( current_device != -1 && current_device != device )   
	{	    
		rc = -EBUSY;
		INFOPRINT(DEVICE_NAME ": device busy!\n");	    
		goto err_out;    
	}    
#else
	device = DEVICE_NR( inode->i_rdev );
	if ( current_device != -1 && current_device != device )
	{
		rc = -EBUSY;
		INFOPRINT(DEVICE_NAME "device busy!\n");
		goto err_out;
	}
#endif
	//paddr = DEFAULT_BANK_MEMORY_ADDRESS;
	//size  = DEFAULT_BANK_LENGTH;
	paddr = AdvSRAMInfo.ul_BankMemoryAddress;
	size  =  AdvSRAMInfo.ul_BankLength  ;
	DEBUGPRINT( DEVICE_NAME": Memory Address is: %lx, Size: %lx\n", paddr, size);		
	uc_ptaddr = ioremap_nocache( paddr, size );  
	//uc_ptaddr = phys_to_virt(paddr);
	DEBUGPRINT( DEVICE_NAME": vaddr = %p\n", uc_ptaddr );		
	
	if ( is_pci == 0 )
	{
		// reset bank number
		outb( 0, AdvSRAMInfo.ul_BankSwitchAddress);  
		// enable SRAM R/W capability  
		outb( 1, AdvSRAMInfo.ul_BankEnableAddress);
	}
	// check SRAM connected or not
	uc_temp = uc_ptaddr[ 0 ];
	uc_ptaddr[ 0 ] = ~uc_ptaddr[ 0 ];
	DEBUGPRINT( DEVICE_NAME": uc_temp = %x, uc_ptaddr[0] = %x \n", uc_temp, uc_ptaddr[ 0 ]);
	
	if( uc_ptaddr[ 0 ] != (unsigned char) (~uc_temp)) 
	{
		INFOPRINT( DEVICE_NAME": SRAM doesn't connect yet! \n");
		if ( is_pci == 0 )
		{
			outb(0, AdvSRAMInfo.ul_BankEnableAddress);   
		}
		iounmap(( void* )uc_ptaddr );      
		rc = -ENODEV;
	}
	uc_ptaddr[ 0 ] = uc_temp; 
	 
	if ( is_pci == 0 )
	{
		outb( 0, AdvSRAMInfo.ul_BankSwitchAddress );
		outb( 0, AdvSRAMInfo.ul_BankEnableAddress );
	}
	return 0;
	
err_out:    
	return rc;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,31)
static int advsram_ioctl(struct block_device *bdev, fmode_t mode,unsigned int cmd, unsigned long arg)
#else
static int advsram_ioctl(struct inode *i, struct file *f, unsigned int cmd, unsigned long arg)
#endif
{
   //if (cmd == HDIO_GETGEO) {
   struct hd_geometry geo;
   switch ( cmd )
   {
	case HDIO_GETGEO:
		geo.heads     = 32;
		geo.sectors   = 2;
		geo.cylinders = (AdvSRAMInfo.ul_DiskSize >> 9) / (geo.heads * geo.sectors);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
		geo.start     = 0;

		if ( copy_to_user((void* /*__user **/)arg, &geo, sizeof(geo)))

#else
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,31)
		geo.start     = get_start_sect(bdev);
#else
		geo.start     = get_start_sect(i->i_bdev);
#endif    
		if (copy_to_user((void __user *) arg, &geo, sizeof(geo)))
#endif
			return -EFAULT;
		break;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	case BLKGETSIZE:
		if (!arg)
			break;
		if ( put_user( asd_sizes[0] <<1, (unsigned long *)arg ))
			return -EFAULT;
		break;
	case BLKGETSIZE64:
		if ( put_user((u64)asd_sizes[0] << 10, (u64*)arg))
			return -EFAULT;
		break;
	case BLKROSET:
	case BLKROGET:
	case BLKSSZGET:
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,31)
		if ( blk_ioctl( bdev, mode,cmd, arg ))
#else
		if ( blk_ioctl( i->i_rdev, cmd, arg ))
#endif
			return -EFAULT;
		break;
#endif
	}
//	return -EINVAL;
	return 0;
}

//=============================================================================
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,31)
static int advsram_release ( struct inode *inode, struct file *filp )
#else
static int advsram_release(struct gendisk *disk, fmode_t mode)
#endif
{
	if ( current_device == -1 )
		return 0;     
	
	/*
	* FIXME: unmap memory
	*/
	
	return 0;
}

// The block device operations defined in fs.h
static struct block_device_operations asd_bdops =
{
	owner:		THIS_MODULE,   
	open:		advsram_open,
	release:	advsram_release,
	ioctl:		advsram_ioctl,
};

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
static struct kobject *sram_find(dev_t dev, int *part, void *data)
{
	*part = 0;
	return get_disk(rd_disk);
}
#endif
static int __init advsram_init ( void )
{
	int ret = -EINVAL, result;
	unsigned int i_ManuID, i_BankSwitch, i_BankNum;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
  	int i;
#endif
	INFOPRINT("\n=========================================================\n");
	INFOPRINT("Advantech SRAM Device Driver. V%s [%s]\n",
		ADVANTECH_SRAM_VER, ADVANTECH_SRAM_DATE);
	INFOPRINT("Supports: Linux block driver for SRAM\n");
	INFOPRINT("Advantech eAutomation Division.\n"); 
	INFOPRINT("=========================================================\n");
	
	// Register major, and accept a dynamic number
	advsd_major = ASD_MAJOR;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)	
	result = register_blkdev(advsd_major, DEVICE_NAME);
#else	
	result = register_blkdev(advsd_major, DEVICE_NAME, &asd_bdops);	
#endif	
	if ( result < 0 ) 
	{
		INFOPRINT( DEVICE_NAME ": can't get major %d\n", advsd_major);
		goto err;
	}
	
	if ( 0 == advsd_major ) 
	{
		advsd_major = result;    //dynamic
	}
   	if ( init_advsram_pci ())
   	{
		return -ENODEV;
   	}
	if(advdev)	//pci device exists
	{
		is_pci = 1;
	}
  	else	//for isa dio
  	{
           i_iobase  = 0x1E0;
           i_iorange = 8;
	   AdvSRAMInfo.ul_BankEnableAddress    = DEFAULT_BANK_ENABLE_ADDRESS;
	   AdvSRAMInfo.ul_BankLength           = DEFAULT_BANK_LENGTH;
	   AdvSRAMInfo.ul_BankMemoryAddress    = DEFAULT_BANK_MEMORY_ADDRESS;
	   AdvSRAMInfo.ul_BankNumber           = DEFAULT_BANK_NUMBER;
	   AdvSRAMInfo.ul_BankNumberAddress    = DEFAULT_BANK_NUMBER_ADDRESS;
	   AdvSRAMInfo.ul_BankSwitchAddress    = DEFAULT_BANK_SWITCH_ADDRESS;
	   AdvSRAMInfo.ul_ManufactureIDAddress = DEFAULT_MANUFACTUREID_ADDRESS;
	
	   INFOPRINT( DEVICE_NAME ": Major Device Number is %d\n", advsd_major );
	
	   // request resources =====
	   if( !request_region( i_iobase, i_iorange, "Advantech SRAMDISK" ))
	   {
		   ret = -EBUSY;
		   INFOPRINT( "Cannot request IO region!\n");	
		   goto out_blkdev;
	   }
	       
	   i_ManuID = inb( i_iobase + 6 );
	   i_BankSwitch = inb( i_iobase );    
	
	   if  ( i_ManuID == 0xAA )
	   {
		   DEBUGPRINT( DEVICE_NAME ": Manufacture ID is 0x%x\n", i_ManuID );
		   i_BankNum = inb( i_iobase + 7 ) + 1;        
		   AdvSRAMInfo.ul_BankNumber = i_BankNum;
		   AdvSRAMInfo.ul_DiskSize = DEFAULT_BANK_LENGTH * i_BankNum;	 
	   }
	   else
	   {
		   DEBUGPRINT( DEVICE_NAME ": Unknown Manufacture ID 0x%x\n", i_ManuID );
		   AdvSRAMInfo.ul_DiskSize  = DEFAULT_BANK_LENGTH * DEFAULT_BANK_NUMBER;
	   }
   	}
	AdvSRAMInfo.ul_DiskSize  = AdvSRAMInfo.ul_BankLength  * AdvSRAMInfo.ul_BankNumber;
	INFOPRINT( DEVICE_NAME ": SRAM size is %ld KBytes\n", 
		ADVSRAM_CHUNK1024(AdvSRAMInfo.ul_DiskSize ));

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	for (  i = 0; i < NUM_SRAMDISKS; i++)
	{
		asd_blocksizes[ i ] = 1024;
		asd_hardsec[ i ]    = 512;
		asd_sizes[ i ]      = ADVSRAM_CHUNK1024(AdvSRAMInfo.ul_DiskSize) ;
	}

	blk_init_queue( BLK_DEFAULT_QUEUE( advsd_major ), do_asd_request );
	hardsect_size[advsd_major]  = asd_hardsec;
	blksize_size[ advsd_major ] = asd_blocksizes;
	blk_size[ advsd_major ]     = asd_sizes;   // in KB
#else
	rd_disk = alloc_disk(1);
	if (!rd_disk)
	{
		ret = -ENOMEM;
		INFOPRINT( DEVICE_NAME ": Error to allocate disk!\n");	
		goto out_region;
	}
	
	rd_queue = blk_init_queue(do_asd_request, &sram_lock);
	if (!rd_queue)
	{
		ret = -ENOMEM;
		INFOPRINT( DEVICE_NAME ": Error to initialize queue!\n");
		goto out_disk;
	}
	
	rd_disk->major = advsd_major;
	rd_disk->first_minor = 0;
	rd_disk->fops = &asd_bdops;
	strcpy(rd_disk->disk_name, "advsram");
	//WangMao add the line below @07/05/10
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
	strcpy(rd_disk->devfs_name, rd_disk->disk_name);
#endif
	//WangMao add end
	set_capacity(rd_disk, AdvSRAMInfo.ul_DiskSize >> 9); // 512 bytes per sector
	rd_disk->queue = rd_queue;
	add_disk(rd_disk);
	blk_register_region(MKDEV(advsd_major, 0), 1, THIS_MODULE,
				sram_find, NULL, NULL);
	return 0;
out_disk:
	put_disk(rd_disk);
out_region:
	release_region( i_iobase, i_iorange );	
#endif
	return 0;
out_blkdev:
	unregister_blkdev(advsd_major, DEVICE_NAME);
err:
	return ret;
}

static void __exit advsram_cleanup( void )
{
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
	blk_unregister_region(MKDEV(advsd_major, 0), 256);
#endif	
	exit_advsram_pci();
	unregister_blkdev( advsd_major, DEVICE_NAME ); 
	//INFOPRINT( DEVICE_NAME ": failed to unregister block device!\n");

	release_region( i_iobase, i_iorange );	
	iounmap(( void* )uc_ptaddr );      

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
	del_gendisk(rd_disk);
	put_disk(rd_disk);
	blk_cleanup_queue( rd_queue);
#else
	blk_cleanup_queue( BLK_DEFAULT_QUEUE( advsd_major));
#endif	
}

module_init(advsram_init);
module_exit(advsram_cleanup); 

#endif
